
{*******************************************************}
{                                                       }
{       RichView                                        }
{       Simple scripting demo, Unicode version          }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

{
  v1.2

  This demo supports:
  1) Text fields, may be multiline: PB_REF, PB_STATUS, PB_DESCRIPTION.
  2) If and If-Not commands, comparing the field value with the constant.

  Not implemented:
  - nested Ifs
  - script error checking (some diagnostic messages in case of wrong command orders, etc.)
  - each command (including < and >) must be inside one item (i.e. all the command's
    characters must be of the same font)
  - (limitation, in case of using tables) Ifs must start and end in the same cell

  All used methods are documented in the help file, except for some code in
  InsertMultilineText. You can simply copy this procedure to your application
  without any modification.

  2011-Oct-11
}

unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, RichView, RVScroll, RVEdit, RVStyle,
  CRVFData, RVItem, RVTable, RVNormalize, RVUni, RVTypes;

type

 TScriptCommandType = (sctBegin, sctEnd, sctIfEqual, sctIfNotEqual, sctField,
   sctConstant);
 TParserState = (psNormal, psNeedField, psNeedConstant, psNeedBegin,
   psSkipTillEnd, psAddTillEnd);

 TParserInfo = record
   State: TParserState;
   Negation: Boolean;
   FieldValue, ConstValue: TRVUnicodeString;
 end;

  TForm1 = class(TForm)
    rve: TRichViewEdit;
    Label1: TLabel;
    Label2: TLabel;
    rv: TRichView;
    GroupBox1: TGroupBox;
    Label3: TLabel;
    txtStatus: TEdit;
    Label4: TLabel;
    txtDescr: TMemo;
    Label5: TLabel;
    txtRef: TEdit;
    RVStyle1: TRVStyle;
    Button1: TButton;
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    { Private declarations }
    function GetCommandType(const Code: TRVUnicodeString): TScriptCommandType;
    function GetFieldValue(const Code: TRVUnicodeString): TRVUnicodeString;
    function GetConstValue(const Code: TRVUnicodeString): TRVUnicodeString;
    procedure ParseDoc(RVData: TCustomRVFormattedData);
    procedure ParseText(RVData: TCustomRVFormattedData; var ItemNo: Integer;
      var Info: TParserInfo);
    procedure InsertMultilineText(RVData: TCustomRVFormattedData;
      var ItemNo, Pos: Integer; const Value: TRVUnicodeString);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses CRVData;

{$R *.dfm}

{ Making sample script }
procedure TForm1.FormCreate(Sender: TObject);
begin
  rve.Clear;
  rve.AddNL('Hello,', 0,0);
  rve.AddNL('', 0,0);
  rve.AddNL('Your ticket is assigned to case number <PB_REF>.', 0,0);
  rve.AddNL('<If Equal><PB_STATUS><Value:Closed><Begin>Case is closed<End>', 0,0);
  rve.AddNL('<If Not Equal><PB_DESCRIPTION><Value:><Begin>', 0,0);
  rve.AddNL('DESCRIPTION:', 0,0);
  rve.AddNL('<PB_DESCRIPTION>', 1,0);
  rve.AddNL('<End>', 0,0);
  rve.AddNL('Thanks for requiring support.', 0,0);
  rve.AddNL('The Helpdesk Team', 0,0);
  rve.Format;
end;
{------------------------------------------------------------------------------}
{ Returns the type of the command inside <>. Code is a text between <> }
function TForm1.GetCommandType(const Code: TRVUnicodeString): TScriptCommandType;
begin
  if Code='Begin' then
    Result := sctBegin
  else if Code='End' then
    Result := sctEnd
  else if Code='If Equal' then
    Result := sctIfEqual
  else if Code='If Not Equal' then
    Result := sctIfNotEqual
  else if Copy(Code, 1, 6)='Value:' then
    Result := sctConstant
  else
    Result := sctField;
end;
{------------------------------------------------------------------------------}
{ Returns the value of the field named Code. Code is a text between <> }
function TForm1.GetFieldValue(const Code: TRVUnicodeString): TRVUnicodeString;
begin
  if Code='PB_REF' then
    Result := txtRef.Text
  else if Code='PB_STATUS' then
    Result := txtStatus.Text
  else if Code='PB_DESCRIPTION' then
    Result := txtDescr.Text
  else
    Result := '!!!UNKNOWN FIELD NAME!!!';
end;
{------------------------------------------------------------------------------}
{ Returns the value of the constant. Code is a text between <>, must be in form
  'Value:'+VALUE }
function TForm1.GetConstValue(const Code: TRVUnicodeString): TRVUnicodeString;
begin
  if Copy(Code, 1, 6)<>'Value:' then
    Result := '!!!INVALID CONSTANT!!!'
  else
    Result := Copy(Code, 7, Length(Code));
end;
{------------------------------------------------------------------------------}
{ Delete item }
procedure SafeDeleteItem(RVData: TCustomRVData; var ItemNo: Integer);
var  BR, PB, PS: Boolean;
begin
  PS := RVData.IsParaStart(ItemNo);
  BR := RVData.IsFromNewLine(ItemNo) and not PS;
  PB := RVData.PageBreaksBeforeItems[ItemNo];
  RVData.DeleteItems(ItemNo, 1);
  dec(ItemNo);
  if (ItemNo<RVData.ItemCount) and (ItemNo>=0) then begin
    RVData.GetItem(ItemNo).SameAsPrev := not PS;
    RVData.GetItem(ItemNo).BR := BR;
    RVData.GetItem(ItemNo).PageBreakBefore := PB;
  end;
end;
{------------------------------------------------------------------------------}
{ Parsing RVData (main document or table cell) }
procedure TForm1.ParseDoc(RVData: TCustomRVFormattedData);
var i,r,c: Integer;
    table: TRVTableItemInfo;
    Info: TParserInfo;
begin
  Info.State :=  psNormal;
  i := 0;
  while i<RVData.ItemCount do begin
    if RVData.GetItemStyle(i)=rvsTable then begin
      table := TRVTableItemInfo(RVData.GetItem(i));
      for r := 0 to table.Rows.Count-1 do
        for c := 0 to table.Rows[r].Count-1 do
          if table.Cells[r,c]<>nil then
            ParseDoc(TCustomRVFormattedData(table.Cells[r,c].GetRVData))
      end
    else if RVData.GetItemStyle(i)>=0 then
      ParseText(RVData, i, Info)
    else if Info.State=psSkipTillEnd then
      SafeDeleteItem(RVData, i);
    inc(i);
  end;
end;
{------------------------------------------------------------------------------}
procedure DeleteW(var s: TRVUnicodeString; Index, Count: Integer);
var s1,s2: TRVUnicodeString;
begin
  if Index-1>0 then
    s1 := Copy(s, 1, Index-1)
  else
    s1 := '';
  if Length(s)-(Index+Count)+1>0 then
    s2 := Copy(s, Index+Count, Length(s)-(Index+Count)+1)
  else
    s2 := '';
  s := s1;
  s := s+s2;
end;
{------------------------------------------------------------------------------}
{ Parsing one text item (the ItemNo-th item of RVData) }
procedure TForm1.ParseText(RVData: TCustomRVFormattedData;
  var ItemNo: Integer; var Info: TParserInfo);
var i, CodeStartPos, SkipStartPos: Integer;
    s, Code: TRVUnicodeString;
    Command: TScriptCommandType;
    Modified: Boolean;
begin
  s := RVData.GetItemTextW(ItemNo);
  CodeStartPos := 0;
  SkipStartPos := 1;
  i := 1;
  Modified := False;
  while i<=Length(s) do begin
    case s[i] of
      '<':
        CodeStartPos := i;
      '>':
        if CodeStartPos>0 then begin
          Code := Copy(s, CodeStartPos+1, i-CodeStartPos-1);
          DeleteW(s, CodeStartPos, i-CodeStartPos+1);
          Modified := True;
          i := CodeStartPos;
          CodeStartPos := -1;
          Command := GetCommandType(Code);
          if Info.State<>psSkipTillEnd then
            case Command of
              sctIfEqual:
                begin
                  Info.State := psNeedField;
                  Info.Negation := False;
                  continue;
                end;
              sctIfNotEqual:
                begin
                  Info.State := psNeedField;
                  Info.Negation := True;
                  continue;
                end;
            end;
          case Info.State of
            psNeedField:
              if Command=sctField then begin
                Info.FieldValue := GetFieldValue(Code);
                Info.State := psNeedConstant;
                continue;
                end
              else
                Info.State := psNormal;
            psNeedConstant:
              if Command=sctConstant then begin
                Info.ConstValue := GetConstValue(Code);
                Info.State := psNeedBegin;
                continue;
                end
              else
                Info.State := psNormal;
            psNeedBegin:
              if Command=sctBegin then begin
                if ((Info.FieldValue=Info.ConstValue) and not Info.Negation) or
                   ((Info.FieldValue<>Info.ConstValue) and Info.Negation) then
                  Info.State := psNormal
                else begin
                  Info.State := psSkipTillEnd;
                  SkipStartPos := i;
                end;
                end
              else
                Info.State := psNormal;
            psSkipTillEnd:
              if Command=sctEnd then begin
                Info.State := psNormal;
                DeleteW(s, SkipStartPos, i-SkipStartPos+1);
                continue;
              end;
            psNormal:
              begin
                if Command=sctEnd then
                  continue;
                if Command=sctField then begin
                  if Modified then
                    RVData.SetItemTextW(ItemNo, s);
                  InsertMultilineText(RVData, ItemNo, i, GetFieldValue(Code));
                  s := RVData.GetItemTextW(ItemNo);
                  Modified := True;
                end;
              end;
          end;
        end;
    end;
    inc(i);
  end;
  if Info.State=psSkipTillEnd then begin
    DeleteW(s, SkipStartPos, Length(s));
    Modified := True;
  end;
  if Modified then begin
    if (s='') and (Info.State=psSkipTillEnd) and (RVData.ItemCount<>1) and
      ((ItemNo=0) or (RVData.GetItemStyle(ItemNo-1)<>rvsListMarker)) then
      SafeDeleteItem(RVData, ItemNo)
    else
      RVData.SetItemTextW(ItemNo, s);
  end;
end;
{------------------------------------------------------------------------------}
type TUnicodeStringDynArray = array of TRVUnicodeString;
{ Assigns lines from s to Lines. It's assumed that line breaks are CR+LF
  characters }
procedure ExtractLines(const s: TRVUnicodeString;
  var Lines: TUnicodeStringDynArray);
var i,j, Count: Integer;
begin
   Count := 1;
   for i := 1 to Length(s) do
     if s[i]=#13 then
       inc(Count);
   SetLength(Lines, Count);
   j := 1;
   Count := 0;
   for i := 1 to Length(s) do
     if s[i]=#13 then begin
       Lines[Count] := Copy(s, j, i-j);
       j := i+2;
       inc(Count);
     end;
   Lines[Count] := Copy(s, j, Length(s));
end;
{------------------------------------------------------------------------------}
{ Inserting Value in the ItemNo-th item of RVData, at position Pos.
  Since Value can be multiline, this procedure may add several items. }
procedure TForm1.InsertMultilineText(RVData: TCustomRVFormattedData;
  var ItemNo, Pos: Integer; const Value: TRVUnicodeString);
var item: TRVTextItemInfo;
    s1, s2: TRVUnicodeString;
    str: TRVRawByteString;
    Lines:  TUnicodeStringDynArray;
    i: Integer;
begin
  if Value='' then
    exit;
  s1 := RVData.GetItemTextW(ItemNo);
  s2 := Copy(s1, Pos, Length(s1));
  s1 := Copy(s1, 1, Pos-1);
  ExtractLines(Value, Lines);
  if High(Lines)=0 then begin
    RVData.SetItemTextW(ItemNo, s1+Value+s2);
    inc(Pos, Length(Value)-1);
    exit;
  end;
  RVData.SetItemTextW(ItemNo, s1+Lines[0]);
  for i := 1 to High(Lines) do begin
    item := RichViewTextItemClass.Create(RVData);
    s1 := Lines[i];
    if i=High(Lines) then
      s1 := s1+s2;
    item.StyleNo := RVData.GetItemStyle(ItemNo);
    item.ParaNo := RVData.GetItemPara(ItemNo);
    str := RVU_GetRawUnicode(s1);
    if RVData.GetRVStyle.TextStyles[item.StyleNo].Unicode then
      item.ItemOptions := item.ItemOptions+[rvioUnicode]
    else
      str := RVU_UnicodeToAnsi(RVData.GetStyleCodePage(item.StyleNo), str);
    item.Inserting(RVData, str, False);
    RVData.Items.InsertObject(ItemNo+i, str, item);
    item.Inserted(RVData, ItemNo+i);
  end;
  ItemNo := ItemNo+High(Lines);
  Pos := Length(Lines[High(Lines)])+1;
end;
{------------------------------------------------------------------------------}
{ Running parser }
procedure TForm1.Button1Click(Sender: TObject);
var Stream: TMemoryStream;
begin
  // copying from rve to rv
  Stream := TMemoryStream.Create;
  rve.SaveRVFToStream(Stream, False);
  Stream.Position := 0;
  rv.LoadRVFFromStream(Stream);
  Stream.Free;
  // parsing
  ParseDoc(rv.RVData);
  // normalizing
  NormalizeRichView(rv.RVData);
  // formatting
  rv.Format;
end;

end.
